#include "cg_user.h"
#include "cg_descent.h"

/*     ________________________________________________________________
      |      A conjugate gradient method with guaranteed descent       |
      |             C-code Version 1.1  (October 6, 2005)              |
      |                    Version 1.2  (November 14, 2005)            |
      |                    Version 2.0  (September 23, 2007)           |
      |           William W. Hager    and   Hongchao Zhang             |
      |          hager@math.ufl.edu       hzhang@math.ufl.edu          |
      |                   Department of Mathematics                    |
      |                     University of Florida                      |
      |                 Gainesville, Florida 32611 USA                 |
      |                      352-392-0281 x 244                        |
      |                                                                |
      |                 Copyright by William W. Hager                  |
      |                                                                |
      |          http://www.math.ufl.edu/~hager/papers/CG              |
      |________________________________________________________________|
       ________________________________________________________________
      |This program is free software; you can redistribute it and/or   |
      |modify it under the terms of the GNU General Public License as  |
      |published by the Free Software Foundation; either version 2 of  |
      |the License, or (at your option) any later version.             |
      |This program is distributed in the hope that it will be useful, |
      |but WITHOUT ANY WARRANTY; without even the implied warranty of  |
      |MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the   |
      |GNU General Public License for more details.                    |
      |                                                                |
      |You should have received a copy of the GNU General Public       |
      |License along with this program; if not, write to the Free      |
      |Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, |
      |MA  02110-1301  USA                                             |
      |________________________________________________________________|
*/

int cg_descent /*  return:
                      -2 (function value became nan)
                      -1 (starting function value is nan)
                       0 (convergence tolerance satisfied)
                       1 (change in func <= feps*|f|)
                       2 (total iterations exceeded maxit)
                       3 (slope always negative in line search)
                       4 (number secant iterations exceed nsecant)
                       5 (search direction not a descent direction)
                       6 (line search fails in initial interval)
                       7 (line search fails during bisection)
                       8 (line search fails during interval update)
                       9 (debugger is on and the function value increases)*/
(
    double            *x , /* input: starting guess, output: the solution */
    INT                n , /* problem dimension */
    cg_stats      *Stats , /* structure with statistics (see cg_descent.h) */
    cg_parameter  *UParm , /* user parameters, NULL = use default parameters */
    double      grad_tol , /* StopRule = 1: |g|_infty <= max (grad_tol,
                                            StopFac*initial |g|_infty) [default]
                              StopRule = 0: |g|_infty <= grad_tol(1+|f|) */
    double    (*cg_value)  /* user provided routine to return the function */
          (double *, INT), /* value cg_value(x) at x */
    void       (*cg_grad)  /* user provided routine cg_grad (g, x), g is*/
     (double *, double *, INT), /* the gradient at x*/
    double         *work   /* work array with at least 4n elements */
)
{
    double  *d, *g, *xtemp, *gtemp, *d1, *d2, *d3, *d4,
            zero, delta2, eta_sq, Qk, Ck,
            f, ftemp, gnorm, xnorm, gnorm2, dnorm2, denom,
            t, t1, t2, t3, t4, t5, dphi, dphi0, alpha, talpha,
            yk, ykyk, ykgk, dkyk, yk1, yk2, yk3, yk4, yk5, beta, tol ;

    int     status, StopRule, cg_tol (double, double, int, double) ;

    INT     n5, iter, maxit, nrestart, i ;

    cg_parameter *Parm, ParmStruc ;
    cg_com Com ;

/* initialize the parameters */

    if ( UParm == NULL )
    {
        Parm = &ParmStruc ;
        cg_default (Parm) ;
    }
    else Parm = UParm ;
    Com.Parm = Parm ;

    if ( Parm->PrintParms )
    {
        printf ("PARAMETERS:\n") ;
        printf ("\n") ;
        printf ("Wolfe line search parameter ..................... delta: %e\n",
                 Parm->delta) ;
        printf ("Wolfe line search parameter ..................... sigma: %e\n",
                 Parm->sigma) ;
        printf ("decay factor for bracketing interval ............ gamma: %e\n",
                 Parm->gamma) ;
        printf ("growth factor for bracket interval ................ rho: %e\n",
                 Parm->rho) ;
        printf ("growth factor for bracket interval after nan .. nan_rho: %e\n",
                 Parm->nan_rho) ;
        printf ("truncation factor for cg beta ..................... eta: %e\n",
                 Parm->eta) ;
        printf ("perturbation parameter for fpert .................. eps: %e\n",
                 Parm->eps) ;
        printf ("factor for computing average cost .............. Qdecay: %e\n",
                 Parm->Qdecay) ;
        printf ("relative change in cost to stop quadstep ... QuadCufOff: %e\n",
                 Parm->QuadCutOff) ;
        printf ("factor multiplying gradient in stop condition . StopFac: %e\n",
                 Parm->StopFac) ;
        printf ("cost change factor, approx Wolfe transition . AWolfeFac: %e\n",
                 Parm->AWolfeFac) ;
        printf ("restart cg every restart_fac*n iterations . restart_fac: %e\n",
                 Parm->restart_fac) ;
        printf ("stop when cost change <= feps*|f| ................. eps: %e\n",
                 Parm->eps) ;
        printf ("starting guess parameter in first iteration ...... psi0: %e\n",
                 Parm->psi0) ;
        printf ("starting step in first iteration if nonzero ...... step: %e\n",
                 Parm->step) ;
        printf ("factor multiply starting guess in quad step ...... psi1: %e\n",
                 Parm->psi1) ;
        printf ("initial guess factor for general iteration ....... psi2: %e\n",
                 Parm->psi2) ;
        printf ("max iterations is n*maxit_fac ............... maxit_fac: %e\n",
                 Parm->maxit_fac) ;
        printf ("max expansions in line search ................. nexpand: %i\n",
                 Parm->nexpand) ;
        printf ("max secant iterations in line search .......... nsecant: %i\n",
                 Parm->nsecant) ;
        printf ("Logical parameters:\n") ;
        if ( Parm->PertRule )
            printf ("    Error estimate for function value is eps\n") ;
        else
            printf ("    Error estimate for function value is eps*Ck\n") ;
        if ( Parm->QuadStep )
            printf ("    Use quadratic interpolation step\n") ;
        else
            printf ("    No quadratic interpolation step\n") ;
        if ( Parm->PrintFinal )
            printf ("    Print final cost and statistics\n") ;
        else
            printf ("    Do not print final cost and statistics\n") ;
        if ( Parm->PrintSteps )
            printf ("    Print each iteration\n") ;
        else
            printf ("    No intermediate printout of iterations\n") ;
        if ( Parm->PrintParms )
            printf ("    Print the parameter structure\n") ;
        else
            printf ("    Do not print parameter structure\n") ;
        if ( Parm->StopRule )
            printf ("    Stopping condition uses initial grad tolerance\n") ;
        else
            printf ("    Stopping condition weighted by absolute cost\n") ;
        if ( Parm->AWolfe)
            printf ("    Approximate Wolfe line search\n") ;
        else
            printf ("    Wolfe line search") ;
            if ( Parm->AWolfeFac > 0. )
                printf (" ... switching to approximate Wolfe\n") ;
            else
                printf ("\n") ;
        if ( Parm->debug)
            printf ("    Check for decay of cost, debugger is on\n") ;
        else
            printf ("    Do not check for decay of cost, debugger is off\n") ;
    }

    zero = 0. ;
    Com.n = n ;          /* problem dimension */
    Com.nf = (INT) 0 ;   /* number of function evaluations */
    Com.ng = (INT) 0 ;   /* number of gradient evaluations */
    Com.AWolfe = Parm->AWolfe ; /* do not touch user's AWolfe */
    StopRule = Parm->StopRule ;

    /* the conjugate gradient algorithm is restarted every nrestart iteration */
    nrestart = (INT) (((double) n)*Parm->restart_fac) ;

    /* abort when number of iterations reaches maxit */
    if ( Parm->maxit_fac == INF ) maxit = INT_INF ;
    else                          maxit = (INT) (((double) n)*Parm->maxit_fac) ;
    
    iter = (INT) 0 ;
    f = zero ;
    n5 = n % 5 ;
    d = work ;
    d1 = d - 1 ;
    d2 = d - 2 ;
    d3 = d - 3 ;
    d4 = d - 4 ;
    g = d+n ;
    xtemp = g+n ;
    gtemp = xtemp+n ;

    Ck = zero ;
    Qk = zero ;

/* initial function and gradient evaluations, initial direction */

    f = cg_value (x, n) ;
    Com.nf++ ;
    cg_grad (g, x, n) ;
    Com.ng++ ;
    Com.f0 = f + f ;
    gnorm = zero ;
    xnorm = zero ;
    gnorm2 = zero ;
    for (i = 0; i < n5; i++)
    {
        t = fabs (x [i]) ;
        xnorm = MAX (xnorm, t) ;
        t = g [i] ;
        d [i] = -t ;
        t = fabs (t) ;
        gnorm2 += t*t ;
        gnorm = MAX (gnorm, t) ;
    }
    for (; i < n;)
    {
        t1 = fabs (x [i]) ;
        xnorm = MAX (xnorm, t1) ;
        t1 = g [i] ;
        d [i] = -t1 ;
        t1 = fabs (t1) ;
        i++ ;

        t2 = fabs (x [i]) ;
        xnorm = MAX (xnorm, t2) ;
        t2 = g [i] ;
        d [i] = -t2 ;
        t2 = fabs (t2) ;
        i++ ;

        t3 = fabs (x [i]) ;
        xnorm = MAX (xnorm, t3) ;
        t3 = g [i] ;
        d [i] = -t3 ;
        t3 = fabs (t3) ;
        i++ ;

        t4 = fabs (x [i]) ;
        xnorm = MAX (xnorm, t4) ;
        t4 = g [i] ;
        d [i] = -t4 ;
        t4 = fabs (t4) ;
        i++ ;

        t5 = fabs (x [i]) ;
        xnorm = MAX (xnorm, t5) ;
        t5 = g [i] ;
        d [i] = -t5 ;
        t5 = fabs (t5) ;
        i++ ;

        gnorm2 += t1*t1 + t2*t2 + t3*t3 + t4*t4 + t5*t5 ;
        t2 = MAX (t2, t1) ;
        t3 = MAX (t3, t2) ;
        t4 = MAX (t4, t3) ;
        t5 = MAX (t5, t4) ;
        gnorm = MAX (gnorm, t5) ;
    }
    /* check that starting function value is nan */
    if ( f != f )
    {
        status = -1 ;
        goto Exit ;
    }

    if ( Parm->StopRule ) tol = MAX (gnorm*Parm->StopFac, grad_tol) ;
    else                  tol = grad_tol ;

    if ( Parm->PrintSteps )
    {
        printf ("iter: %5i f = %14.6e gnorm = %14.6e AWolfe = %2i\n",
          (int) iter, f, gnorm, Com.AWolfe) ;
    }

    if ( cg_tol (f, gnorm, StopRule, tol) )
    {
        status = 0 ;
        goto Exit ;
    }

    dphi0 = -gnorm2 ;
    delta2 = 2*Parm->delta - 1. ;
    eta_sq = Parm->eta*Parm->eta ;
    alpha = Parm->step ;
    if ( alpha == 0. )
    {
        alpha = Parm->psi0*xnorm/gnorm ;
        if ( xnorm == zero )
        {
            if ( f != zero ) alpha = Parm->psi0*fabs (f)/gnorm2 ;
            else             alpha = 1. ;
        }
    }
 
/*  start the conjugate gradient iteration
    alpha starts as old step, ends as final step for current iteration
    f is function value for alpha = 0
    QuadOK = TRUE means that a quadratic step was taken */
 
    for (iter = 1; iter <= maxit; iter++)
    {
        Com.QuadOK = FALSE ;
        alpha = Parm->psi2*alpha ;
        if ( Parm->QuadStep )
        {
            if ( f != zero ) t = fabs ((f-Com.f0)/f) ;
            else             t = 1. ;
            if ( t > Parm->QuadCutOff )       /* take provisional step talpha */
            {
                talpha = Parm->psi1*alpha ;
                cg_step (xtemp, x, d, talpha, n) ;
                ftemp = cg_value (xtemp, n) ; /* provisional function value */
                Com.nf++ ;

                /* check if function value is nan */
                if ( ftemp != ftemp ) /* reduce stepsize */
                {
                    for (i = 0; i < Parm->nexpand; i++)
                    {
                        talpha /= Parm->rho ;
                        cg_step (xtemp, x, d, talpha, n) ;
                        ftemp = cg_value (xtemp, n) ;
                        Com.nf++ ;
                        if ( ftemp == ftemp ) break ;
                    }
                    if ( i == Parm->nexpand )
                    {
                        status = -2 ;
                        goto Exit ;
                    }
                }

                if ( ftemp < f )              /* check if quadstep > 0 */
                {
                   denom = 2.*(((ftemp-f)/talpha)-dphi0) ;
                   if ( denom > zero )        /* try a quadratic fit step */
                   {
                       Com.QuadOK = TRUE ;
                       alpha = -dphi0*talpha/denom ;
                   }
                }
            }
        }
        Com.f0 = f ;                          /* f0 saved as prior value */

        if ( Parm->PrintSteps )
        {
            printf ("QuadOK: %2i initial a: %14.6e f0: %14.6e dphi: %14.6e\n",
                    Com.QuadOK, alpha, Com.f0, dphi0) ;
        }

/* parameters in Wolfe and approximate Wolfe conditions, and in update */

        Qk = Parm->Qdecay*Qk + 1. ;
        Ck = Ck + (fabs (f) - Ck)/Qk ;        /* average cost magnitude */

        if ( Parm->PertRule ) Com.fpert = f + Parm->eps*Ck ;
        else                  Com.fpert = f + Parm->eps ;

        Com.wolfe_hi = Parm->delta*dphi0 ;
        Com.wolfe_lo = Parm->sigma*dphi0 ;
        Com.awolfe_hi = delta2*dphi0 ;
        Com.alpha = alpha ;        /* either prior step or quadratic fit step */
        Com.f = f ;
        Com.df = dphi ;
        
        if ( Com.AWolfe )                     /* approximate Wolfe line search*/
        {
            status =
                cg_line (dphi0, x, xtemp, d, gtemp, cg_value, cg_grad, &Com) ;
        }
        else                                  /* ordinary Wolfe line search */
        {
            status =
                cg_lineW (dphi0, x, xtemp, d, gtemp, cg_value, cg_grad, &Com) ;
        }

        alpha = Com.alpha ;
        f = Com.f ;
        dphi = Com.df ;

        if ( status ) goto Exit ;

/*Test for convergence to within machine epsilon
  [set feps to zero to remove this test] */
 
        if ( -alpha*dphi0 <= Parm->feps*fabs (f) )
        {
            status = 1 ;
            goto Exit ;
        }

/* compute beta, yk2, gnorm, gnorm2, dnorm2, update x and g */

        if ( iter % nrestart != 0 )
        {
            gnorm = zero ;
            dnorm2 = zero ;
            ykyk = zero ;
            ykgk = zero ;
            for (i = 0; i < n5; i++)
            {
                x [i] = xtemp [i] ;
                t = gtemp [i] ;
                yk = t - g [i] ;
                ykyk += yk*yk ;
                ykgk += yk*t ;
                g [i] = t ;
                t = fabs (t) ;
                gnorm = MAX (gnorm, t) ;
                dnorm2 = dnorm2 + d [i]*d [i] ;
            }
            for (; i < n; )
            {
                x [i] = xtemp [i] ;
                t1 = gtemp [i] ;
                yk1 = t1 - g [i] ;
                g [i] = t1 ;
                i++ ;

                x [i] = xtemp [i] ;
                t2 = gtemp [i] ;
                yk2 = t2 - g [i] ;
                g [i] = t2 ;
                i++ ;

                x [i] = xtemp [i] ;
                t3 = gtemp [i] ;
                yk3 = t3 - g [i] ;
                g [i] = t3 ;
                i++ ;

                x [i] = xtemp [i] ;
                t4 = gtemp [i] ;
                yk4 = t4 - g [i] ;
                g [i] = t4 ;
                i++ ;

                x [i] = xtemp [i] ;
                t5 = gtemp [i] ;
                yk5 = t5 - g [i] ;
                g [i] = t5 ;

                dnorm2 = dnorm2 + d [i]*d [i] + d1 [i]*d1 [i] + d2 [i]*d2 [i]
                                              + d3 [i]*d3 [i] + d4 [i]*d4 [i] ;
                i++ ;
                ykyk += yk1*yk1 + yk2*yk2 + yk3*yk3 + yk4*yk4 + yk5*yk5 ;
                ykgk += yk1*t1 + yk2*t2 + yk3*t3 + yk4*t4 + yk5*t5 ;
                t1 = fabs (t1) ;
                gnorm = MAX (gnorm, t1) ;
                t2 = fabs (t2) ;
                gnorm = MAX (gnorm, t2) ;
                t3 = fabs (t3) ;
                gnorm = MAX (gnorm, t3) ;
                t4 = fabs (t4) ;
                gnorm = MAX (gnorm, t4) ;
                t5 = fabs (t5) ;
                gnorm = MAX (gnorm, t5) ;
            }

            if ( cg_tol (f, gnorm, StopRule, tol) )
            {
                status = 0 ;
                goto Exit ;
            }
            dkyk = dphi - dphi0 ;
            beta = (ykgk - 2.*dphi*ykyk/dkyk)/dkyk ;
/*
    faster: initialize dnorm2 = gnorm2 at start, then
            dnorm2 = gnorm2 + beta**2*dnorm2 - 2.*beta*dphi
            gnorm2 = ||g_{k+1}||^2
            dnorm2 = ||d_{k+1}||^2
            dpi = g_{k+1}' d_k */

            t = -1./sqrt (dnorm2*MIN (eta_sq, gnorm2)) ;
            beta = MAX (beta, t) ;

/*    update search direction d = -g + beta*dold */

            gnorm2 = zero ;
            for (i = 0; i < n5; i++)
            {
                t = g [i] ;
                d [i] = -t + beta*d [i] ;
                gnorm2 += t*t ;
            }
            for (; i < n; )
            {
                t1 = g [i] ;
                d [i] = -t1 + beta*d [i] ;
                i++ ;

                t2 = g [i] ;
                d [i] = -t2 + beta*d [i] ;
                i++ ;

                t3 = g [i] ;
                d [i] = -t3 + beta*d [i] ;
                i++ ;

                t4 = g [i] ;
                d [i] = -t4 + beta*d [i] ;
                i++ ;

                t5 = g [i] ;
                d [i] = -t5 + beta*d [i] ;
                i++ ;

                gnorm2 += t1*t1 + t2*t2 + t3*t3 + t4*t4 + t5*t5 ;
            }
            dphi0 = -gnorm2 + beta*dphi ;
        }
        else
        {

/*    search direction d = -g */

            if ( Parm->PrintSteps ) printf ("RESTART CG\n") ;
            gnorm = zero ;
            gnorm2 = zero ;
            for (i = 0; i < n5; i++)
            {
                x [i] = xtemp [i] ;
                t = gtemp [i] ;
                g [i] = t ;
                d [i] = -t ;
                t = fabs (t) ;
                gnorm = MAX (gnorm, t) ;
                gnorm2 += t*t ;
            }
            for (; i < n; )
            {
                x [i] = xtemp [i] ;
                t1 = gtemp [i] ;
                g [i] = t1 ;
                d [i] = -t1 ;
                t1 = fabs (t1) ;
                i++ ;

                x [i] = xtemp [i] ;
                t2 = gtemp [i] ;
                g [i] = t2 ;
                d [i] = -t2 ;
                t2 = fabs (t2) ;
                i++ ;

                x [i] = xtemp [i] ;
                t3 = gtemp [i] ;
                g [i] = t3 ;
                d [i] = -t3 ;
                t3 = fabs (t3) ;
                i++ ;

                x [i] = xtemp [i] ;
                t4 = gtemp [i] ;
                g [i] = t4 ;
                d [i] = -t4 ;
                t4 = fabs (t4) ;
                i++ ;

                x [i] = xtemp [i] ;
                t5 = gtemp [i] ;
                g [i] = t5 ;
                d [i] = -t5 ;
                t5 = fabs (t5) ;
                i++ ;

                t2 = MAX (t2, t1) ;
                t3 = MAX (t3, t2) ;
                t4 = MAX (t4, t3) ;
                t5 = MAX (t5, t4) ;
                gnorm = MAX (gnorm, t5) ;
                gnorm2 += t1*t1 + t2*t2 + t3*t3 + t4*t4 + t5*t5 ;
            }
            if ( cg_tol (f, gnorm, StopRule, tol) )
            {
                status = 0 ;
                goto Exit ;
            }
            dphi0 = -gnorm2 ;
        }
        if ( !Com.AWolfe )
        {
            if ( fabs (f-Com.f0) < Parm->AWolfeFac*Ck ) Com.AWolfe = TRUE ;
        }
    
        if ( Parm->PrintSteps )
        {
            printf ("\niter: %5i f = %14.6e gnorm = %14.6e AWolfe = %2i\n",
               (int) iter, f, gnorm, Com.AWolfe) ;
        }

        if ( Parm->debug )
        {
            if ( f > Com.f0 + 1.e-10*Ck )
            {
                status = 9 ;
                goto Exit ;
            }
        }
                
        if ( dphi0 > zero )
        {
           status = 5 ;
           goto Exit ;
        }
    }
    status = 2 ;

Exit:
    Stats->f = f ;
    Stats->gnorm = gnorm ;
    Stats->nfunc = Com.nf ;
    Stats->ngrad = Com.ng ;
    Stats->iter = iter ;
    if ( status > 2 )
    {
        gnorm = zero ;
        for (i = 0; i < n; i++)
        {
            x [i] = xtemp [i] ;
            g [i] = gtemp [i] ;
            t = fabs (g [i]) ;
            gnorm = MAX (gnorm, t) ;
        }
        Stats->gnorm = gnorm ;
    }
    if ( Parm->PrintFinal || Parm->PrintSteps )
    {
        const char mess1 [] = "Possible causes of this error message:" ;
        const char mess2 [] = "   - your tolerance may be too strict: "
                              "grad_tol = " ;
        const char mess3 [] = "Line search fails" ;
        const char mess4 [] = "   - your gradient routine has an error" ;
        const char mess5 [] = "   - the parameter epsilon in cg_descent_c.parm "
                              "is too small" ;
        printf ("\nTermination status: %i\n", status) ;
        if ( status == -2 )
        {
            printf ("At iteration %i function value became nan\n", (int) iter) ;
        }
        else if ( status == -1 )
        {
            printf ("Objective function value is nan at starting point\n") ;
        }
        else if ( status == 0 )
        {
            printf ("Convergence tolerance for gradient satisfied\n") ;
        }
        else if ( status == 1 )
        {
            printf ("Terminating since change in function value "
                    "<= feps*|f|\n") ;
        }
        else if ( status == 2 )
        {
            printf ("Number of iterations reached max allowed: %15.7g\n",
                (double) maxit) ;
            printf ("%s\n", mess1) ;
            printf ("%s %e\n", mess2, grad_tol) ;
        }
        else if ( status == 3 )
        {
            printf ("Slope always negative in line search\n") ;
            printf ("%s\n", mess1) ;
            printf ("   - your cost function has an error\n") ;
            printf ("%s\n", mess4) ;
        }
        else if ( status == 4 )
        {
            printf ("Line search fails, too many secant steps\n") ;
            printf ("%s\n", mess1) ;
            printf ("%s %e\n", mess2, grad_tol) ;
        }
        else if ( status == 5 )
        {
            printf ("Search direction not a descent direction\n") ;
        }
        else if ( status == 6 )
        {
            printf ("%s\n", mess3) ;
            printf ("%s\n", mess1) ;
            printf ("%s %e\n", mess2, grad_tol) ;
            printf ("%s\n", mess4) ;
            printf ("%s\n", mess5) ;
        }
        else if ( status == 7 )
        {
            printf ("%s\n", mess3) ;
            printf ("%s\n", mess1) ;
            printf ("%s %e\n", mess2, grad_tol) ;
        }
        else if ( status == 8 )
        {
            printf ("%s\n", mess3) ;
            printf ("%s\n", mess1) ;
            printf ("%s %e\n", mess2, grad_tol) ;
            printf ("%s\n", mess4) ;
            printf ("%s\n", mess5) ;
        }
        else if ( status == 9 )
        {
            printf ("Debugger is on, function value does not improve\n") ;
            printf ("new value: %25.16e old value: %25.16e\n", f, Com.f0) ;
        }

        printf ("absolute largest component of gradient: %e\n", gnorm) ;
        printf ("function value:       %15.7e\n", f) ;
        printf ("cg iterations:        %15.7e\n", (double) iter) ;
        printf ("function evaluations: %15.7e\n", (double) Com.nf) ;
        printf ("gradient evaluations: %15.7e\n", (double) Com.ng) ;
    }
    return (status) ;
}

/*
    PARAMETERS:
 
    delta - range [0, .5], used in the Wolfe conditions
    sigma - range [delta, 1], used in the Wolfe conditions
    eps - range [0, infty], used to compute line search perturbation
    gamma - range [0,1], determines when to perform bisection step
    rho   - range [1, infty], growth factor when finding initial interval
    eta   - range [0, infty], used in lower bound for beta
    psi0  - range [0, 1], factor used in very initial starting guess
    psi1  - range [0, 1], factor previous step multiplied by in QuadStep
    psi2  - range [1, infty], factor previous step is multipled by for startup
    step  - range [0, infty], nonzero if initial step of initial line search
    QuadCutOff - perform QuadStep if relative change in f > QuadCutOff
    StopFac - used in StopRule
    AWolfeFac - used to decide when to switch from Wolfe to AWolfe
    restart_fac - range [0, infty] restart cg when iter = n*restart 
    maxit_fac - range [0, infty] terminate in maxit = maxit_fac*n iterations
    feps - stop when -alpha*dphi0 [est. change in value] <= feps*|f|
           [feps = 0 removes this test, example: feps = eps*1.e-5
            where eps is machine epsilon]
    tol   - range [0, infty], convergence tolerance
    nexpand - range [0, infty], number of grow/shrink allowed in bracket
    nsecant - range [0, infty], maximum number of secant steps
    PertRule - gives the rule used for the perturbation in f
                F => fpert = eps
                T => fpert = eps*Ck, Ck is an average of prior |f|
                            Ck is an average of prior |f|
    QuadStep- TRUE [use quadratic step] FALSE [no quadratic step]
    PrintSteps- FALSE [no printout] TRUE [print intermediate results]
    PrintFinal- FALSE [no printout] TRUE [print messages, final error]
    StopRule - TRUE [max abs grad <= max [tol, StopFac*initial abs grad]]
               FALSE [... <= tol*[1+|f|]]
    AWolfe - FALSE [use standard Wolfe initially]
           - TRUE [use approximate + standard Wolfe]
    debug - FALSE [no debugging]
          - TRUE [check that function values do not increase]
*/

/* Set default parameter values. If the parameter argument of
   cg_descent is NULL, this routine is called by cg_descent automatically.
   If the user wishes to set parameter values, then a cg_parameter structure
   could be allocated in the main program, the user could call cg_default
   to initialize the structure, and then individual elements in the structure
   could be changed, before passing the structure to cg_descent. */

void cg_default
(
    cg_parameter   *Parm
)
{
    Parm->delta = .1 ;     /* Wolfe line search parameter */
    Parm->sigma = .9 ;     /* Wolfe line search parameter */
    Parm->gamma = .66 ;    /* decay factor for bracket interval width */
    Parm->rho = 5. ;       /* growth factor when searching for initial
                                bracketing interval */
    Parm->nan_rho = 1.3 ;  /* after encountering nan, growth factor when
                                searching for initial bracketing interval */
    Parm->eta = .01 ;      /* lower bound for the conjugate gradient update
                                parameter beta_k is eta*||d||_2 */
    Parm->eps = 1.e-6 ;    /* perturbation parameter for computing fpert */
    Parm->Qdecay = .7 ;    /* factor used to compute average cost */
    Parm->QuadCutOff = 1.e-12 ; /* QuadStep used when relative change in f >
                                   QuadCutOff */
    Parm->StopFac = 0.e-12 ;  /* factor multiplying starting |grad|_infty in
                                 the stopping condition StopRule (see below) */
    Parm->AWolfeFac = 1.e-3 ; /* if AWolfe is F, then set AWolfe = T when
                                 |f-f0| < Awolfe_fac*Ck) */
    Parm->restart_fac = 1. ;/* restart when iteration = n*restart_fac */
    Parm->feps = 0. ;       /* stop when value change <= feps*|f| */
    Parm->psi0 = .01 ;      /* factor used in starting guess for iteration 1 */
    Parm->psi1 = .1 ;       /* in performing a QuadStep, we evaluate the
                                 function at psi1*previous step */
    Parm->psi2 = 2. ;       /* when starting a new cg iteration, our initial
                                 guess for the line search stepsize is
                                 psi2*previous step */
    Parm->step = 0. ;       /* if nonzero, it is the initial step of initial
                                line search */
    Parm->nexpand = (int) 50 ; /* abort the line search if the bracketing
                                  interval is expanded nexpand times */
    Parm->nsecant = (int) 50 ; /* abort the line search if the number of secant
                                  iterations is nsecant */
    Parm->maxit_fac = INF ;    /* abort when number of iterations reaches maxit
                                  (= maxit_fac*n) */
    
    /*logical parameters */

    Parm->PertRule = TRUE ;    /* estimated error in  function value,
                                  F => eps, T => eps*Ck */
    Parm->QuadStep = TRUE ;    /* T (quadratic interpolation in line search)*/
    Parm->PrintFinal = TRUE ;  /* T (print final value) F (no print) */
    Parm->PrintSteps = FALSE ; /* T (print iteration steps) F (no printout) */
    Parm->PrintParms = FALSE ; /* T (print parameters) F (do not display Parms*/
    Parm->StopRule = TRUE ;    /* T (|grad|_infty <=
                                     max(tol, initial |grad|_infty*StopFact)
                                  F (|grad|_infty <= tol*(1+|f|)) */
    Parm->AWolfe = FALSE ;     /* T (use approximate Wolfe line search)
                                  F (use Wolfe line search) */
    Parm->debug = FALSE ;      /* T (check for no increase in function value)
                                  F (no debugging) */
}

/* Read the parameters from a specified parameter file. The parameters
   should be stored in the same order as used in versions 1.1 and 1.2
   of the C code, or versions 1.1 to 1.4 of the Fortran code. This routine
   is provided for compatibility with earlier versions of cg_descent.
   This is called from the main program where a cg_parameter structure
   has been allocated. You could further modify the elements of the
   parameter structure before passing it to cg_descent */

int  cg_readParms /* return  0 (parameter file was read)
                            -1 (parameter file not found)
                            -2 (missing entry in parameter file)
                            -3 (comment in parameter file too long) */
(
    char        *filename , /* name of file containing parameters */
    cg_parameter    *Parm   /* pointer to a cg parameter structure */
)
{
    char junk [CGMAXLINE+1] ;
    int info ;
    FILE *ParmFile ;

    ParmFile = fopen (filename, "r") ;
    if ( ParmFile == NULL ) return (-1) ;

    info = fscanf (ParmFile, "%lg", &(Parm->delta)) ;
    if ( info != 1 )
    {
        printf ("delta parameter missing in cg_descent parameter file %s\n",
                 filename) ;
        return (-2) ;
    }
    fgets (junk, CGMAXLINE, ParmFile) ;
    if (strlen (junk) >= CGMAXLINE-1) return (-3) ;

    info = fscanf (ParmFile, "%lg", &(Parm->sigma)) ;
    if ( info != 1 )
    {
        printf ("sigma parameter missing in cg_descent parameter file %s\n",
                 filename) ;
        return (-2) ;
    }
    fgets (junk, CGMAXLINE, ParmFile) ;
    if (strlen (junk) >= CGMAXLINE-1) return (-3) ;

    info = fscanf (ParmFile, "%lg", &(Parm->eps)) ;
    if ( info != 1 )
    {
        printf ("eps parameter missing in cg_descent parameter file %s\n",
                 filename) ;
        return (-2) ;
    }
    fgets (junk, CGMAXLINE, ParmFile) ;
    if (strlen (junk) >= CGMAXLINE-1) return (-2) ;

    info = fscanf (ParmFile, "%lg", &(Parm->gamma)) ;
    if ( info != 1 )
    {
        printf ("gamma parameter missing in cg_descent parameter file %s\n",
                 filename) ;
        return (-2) ;
    }
    fgets (junk, CGMAXLINE, ParmFile) ;
    if (strlen (junk) >= CGMAXLINE-1) return (-3) ;

    info = fscanf (ParmFile, "%lg", &(Parm->rho)) ;
    if ( info != 1 )
    {
        printf ("rho parameter missing in cg_descent parameter file %s\n",
                 filename) ;
        return (-2) ;
    }
    fgets (junk, CGMAXLINE, ParmFile) ;
    if (strlen (junk) >= CGMAXLINE-1) return (-3) ;

    info = fscanf (ParmFile, "%lg", &(Parm->eta)) ;
    if ( info != 1 )
    {
        printf ("eta parameter missing in cg_descent parameter file %s\n",
                 filename) ;
        return (-2) ;
    }
    fgets (junk, CGMAXLINE, ParmFile) ;
    if (strlen (junk) >= CGMAXLINE-1) return (-3) ;

    info = fscanf (ParmFile, "%lg", &(Parm->psi0)) ;
    if ( info != 1 )
    {
        printf ("psi0 parameter missing in cg_descent parameter file %s\n",
                 filename) ;
        return (-2) ;
    }
    fgets (junk, CGMAXLINE, ParmFile) ;
    if (strlen (junk) >= CGMAXLINE-1) return (-3) ;

    info = fscanf (ParmFile, "%lg", &(Parm->psi1)) ;
    if ( info != 1 )
    {
        printf ("psi1 parameter missing in cg_descent parameter file %s\n",
                 filename) ;
        return (-2) ;
    }
    fgets (junk, CGMAXLINE, ParmFile) ;
    if (strlen (junk) >= CGMAXLINE-1) return (-3) ;

    info = fscanf (ParmFile, "%lg", &(Parm->psi2)) ;
    if ( info != 1 )
    {
        printf ("psi2 parameter missing in cg_descent parameter file %s\n",
                 filename) ;
        return (-2) ;
    }
    fgets (junk, CGMAXLINE, ParmFile) ;
    if (strlen (junk) >= CGMAXLINE-1) return (-3) ;

    info = fscanf (ParmFile, "%lg", &(Parm->QuadCutOff)) ;
    if ( info != 1 )
    {
        printf ("QuadCutOff parameter missing in cg_descent parameter "
                 "file %s\n", filename) ;
        return (-2) ;
    }
    fgets (junk, CGMAXLINE, ParmFile) ;
    if (strlen (junk) >= CGMAXLINE-1) return (-3) ;

    info = fscanf (ParmFile, "%lg", &(Parm->StopFac)) ;
    if ( info != 1 )
    {
        printf ("StopFact parameter missing in cg_descent parameter file %s\n",
                 filename) ;
        return (-2) ;
    }
    fgets (junk, CGMAXLINE, ParmFile) ;
    if (strlen (junk) >= CGMAXLINE-1) return (-3) ;

    info = fscanf (ParmFile, "%lg", &(Parm->AWolfeFac)) ;
    if ( info != 1 )
    {
        printf ("AWolfeFac parameter missing in cg_descent parameter file %s\n",
                 filename) ;
        return (-2) ;
    }
    fgets (junk, CGMAXLINE, ParmFile) ;
    if (strlen (junk) >= CGMAXLINE-1) return (-3) ;

    info = fscanf (ParmFile, "%lg", &(Parm->restart_fac)) ;
    if ( info != 1 )
    {
        printf ("restart_fac parameter missing in cg_descent parameter "
                 "file %s\n", filename) ;
        return (-2) ;
    }
    fgets (junk, CGMAXLINE, ParmFile) ;
    if (strlen (junk) >= CGMAXLINE-1) return (-3) ;

    info = fscanf (ParmFile, "%lg", &(Parm->maxit_fac)) ;
    if ( info != 1 )
    {
        printf ("maxit_fac parameter missing in cg_descent parameter "
                 "file %s\n", filename) ;
        return (-2) ;
    }
    fgets (junk, CGMAXLINE, ParmFile) ;
    if (strlen (junk) >= CGMAXLINE-1) return (-3) ;

    info = fscanf (ParmFile, "%lg", &(Parm->feps)) ;
    if ( info != 1 )
    {
        printf ("feps parameter missing in cg_descent parameter "
                 "file %s\n", filename) ;
        return (-2) ;
    }
    fgets (junk, CGMAXLINE, ParmFile) ;
    if (strlen (junk) >= CGMAXLINE-1) return (-3) ;

    info = fscanf (ParmFile, "%lg", &(Parm->Qdecay)) ;
    if ( info != 1 )
    {
        printf ("Qdecay parameter missing in cg_descent parameter file %s\n",
                 filename) ;
        return (-2) ;
    }
    fgets (junk, CGMAXLINE, ParmFile) ;
    if (strlen (junk) >= CGMAXLINE-1) return (-3) ;

    info = fscanf (ParmFile, "%i", &(Parm->nexpand)) ;
    if ( info != 1 )
    {
        printf ("nexpand parameter missing in cg_descent parameter file %s\n",
                 filename) ;
        return (-2) ;
    }
    fgets (junk, CGMAXLINE, ParmFile) ;
    if (strlen (junk) >= CGMAXLINE-1) return (-3) ;

    info = fscanf (ParmFile, "%i", &(Parm->nsecant)) ;
    if ( info != 1 )
    {
        printf ("nsecant parameter missing in cg_descent parameter file %s\n",
                 filename) ;
        return (-2) ;
    }
    fgets (junk, CGMAXLINE, ParmFile) ;
    if (strlen (junk) >= CGMAXLINE-1) return (-3) ;

    info = fscanf (ParmFile, "%i", &(Parm->PertRule)) ;
    if ( info != 1 )
    {
        printf ("PertRule parameter missing in cg_descent parameter file %s\n",
                 filename) ;
        return (-2) ;
    }
    fgets (junk, CGMAXLINE, ParmFile) ;
    if (strlen (junk) >= CGMAXLINE-1) return (-3) ;

    info = fscanf (ParmFile, "%i", &(Parm->QuadStep)) ;
    if ( info != 1 )
    {
        printf ("QuadStep parameter missing in cg_descent parameter file %s\n",
                 filename) ;
        return (-2) ;
    }
    fgets (junk, CGMAXLINE, ParmFile) ;
    if (strlen (junk) >= CGMAXLINE-1) return (-3) ;

    info = fscanf (ParmFile, "%i", &(Parm->PrintSteps)) ;
    if ( info != 1 )
    {
        printf ("PrintLevel parameter missing in cg_descent parameter "
                "file %s\n", filename) ;
        return (-2) ;
    }
    fgets (junk, CGMAXLINE, ParmFile) ;
    if (strlen (junk) >= CGMAXLINE-1) return (-3) ;

    info = fscanf (ParmFile, "%i", &(Parm->PrintFinal)) ;
    if ( info != 1 )
    {
        printf ("PrintFinal parameter missing in cg_descent parameter "
                "file %s\n", filename) ;
        return (-2) ;
    }
    fgets (junk, CGMAXLINE, ParmFile) ;
    if (strlen (junk) >= CGMAXLINE-1) return (-3) ;

    info = fscanf (ParmFile, "%i", &(Parm->StopRule)) ;
    if ( info != 1 )
    {
        printf ("StopRule parameter missing in cg_descent parameter "
                "file %s\n", filename) ;
        return (-2) ;
    }
    fgets (junk, CGMAXLINE, ParmFile) ;
    if (strlen (junk) >= CGMAXLINE-1) return (-3) ;

    info = fscanf (ParmFile, "%i", &(Parm->AWolfe)) ;
    if ( info != 1 )
    {
        printf ("AWolfe parameter missing in cg_descent parameter "
                "file %s\n", filename) ;
        return (-2) ;
    }
    fgets (junk, CGMAXLINE, ParmFile) ;
    if (strlen (junk) >= CGMAXLINE-1) return (-3) ;

    info = fscanf (ParmFile, "%lg", &(Parm->step)) ;
    if ( info != 1 )
    {
        printf ("Step parameter missing in cg_descent parameter "
                "file %s\n", filename) ;
        return (-2) ;
    }
    fgets (junk, CGMAXLINE, ParmFile) ;
    if (strlen (junk) >= CGMAXLINE-1) return (-3) ;

    info = fscanf (ParmFile, "%i", &(Parm->debug)) ;
    if ( info != 1 )
    {
        printf ("debug parameter missing in cg_descent parameter "
                "file %s\n", filename) ;
        return (-2) ;
    }
    fgets (junk, CGMAXLINE, ParmFile) ;
    if (strlen (junk) >= CGMAXLINE-1) return (-3) ;
    fclose (ParmFile) ;
    Parm->PrintParms = FALSE ;
    Parm->nan_rho = 1.3 ;
    return (0) ;
}

/*check whether the Wolfe or the approximate Wolfe conditions are satisfied */

int cg_Wolfe
(
    double       alpha , /* stepsize */
    double           f , /* function value associated with stepsize alpha */
    double        dphi , /* derivative value associated with stepsize alpha */
    cg_com        *Com   /* cg com */
)
{
    if ( dphi >= Com->wolfe_lo )
    {

/* test original Wolfe conditions */

        if ( f - Com->f0 <= alpha*Com->wolfe_hi )
        {
            if ( Com->Parm->PrintSteps )
            {
                printf ("wolfe f: %14.6e f0: %14.6e dphi: %14.6e\n",
                         f, Com->f0, dphi) ;
            }
            return (1) ;
        }
/* test approximate Wolfe conditions */
        else if ( Com->AWolfe )
        {
            if ( (f <= Com->fpert) && (dphi <= Com->awolfe_hi) )
            {
                if ( Com->Parm->PrintSteps )
                {
                    printf ("f: %14.6e fpert: %14.6e dphi: %14.6e awolf_hi: "
                            "%14.6e\n", f, Com->fpert, dphi, Com->awolfe_hi) ;
                }
                return (1) ;
            }
        }
    }
    return (0) ;
}

/* check for convergence of the cg iterations */

int cg_tol
(
    double           f , /* function value associated with stepsize */
    double       gnorm , /* gradient sup-norm */
    int       StopRule , /* T => |grad|_infty <=max (tol, |grad|_infty*StopFact)
                            F => |grad|_infty <= tol*(1+|f|)) */
    double         tol   /* tolerance */
)
{
    if ( StopRule )
    {
        if ( gnorm <= tol ) return (1) ;
    }
    else if ( gnorm <= tol*(1.0 + fabs (f)) ) return (1) ;
    return (0) ;
}

/* compute dot product of x and y, vectors of length n */

double cg_dot
(
    double *x , /* first vector */
    double *y , /* second vector */
    INT     n  /* length of vectors */
)
{
    INT i, n5 ;
    double t ;
    t = 0. ;
    n5 = n % 5 ;
    for (i = 0; i < n5; i++) t += x [i]*y [i] ;
    for (; i < n; i += 5)
    {
        t += x [i]*y[i] + x [i+1]*y [i+1] + x [i+2]*y [i+2]
           + x [i+3]*y [i+3] + x [i+4]*y [i+4] ;
    }
    return (t) ;
}

/* compute xtemp = x + alpha d */

void cg_step
(
    double *xtemp , /*output vector */
    double     *x , /* initial vector */
    double     *d , /* search direction */
    double  alpha , /* stepsize */
    INT         n   /* length of the vectors */
)
{
    INT n5, i ;
    n5 = n % 5 ;
    for (i = 0; i < n5; i++) xtemp [i] = x[i] + alpha*d[i] ;
    for (; i < n;)
    { 
        xtemp [i] = x [i] + alpha*d [i] ;
        i++ ;
        xtemp [i] = x [i] + alpha*d [i] ;
        i++ ;
        xtemp [i] = x [i] + alpha*d [i] ;
        i++ ;
        xtemp [i] = x [i] + alpha*d [i] ;
        i++ ;
        xtemp [i] = x [i] + alpha*d [i] ;
        i++ ;
    }
}

/* approximate Wolfe line search routine */

int cg_line
(
    double       dphi0 , /* function derivative at starting point (alpha = 0) */
    double          *x , /* current iterate */
    double      *xtemp , /* x + alpha*d */
    double          *d , /* current search direction */
    double      *gtemp , /* gradient at x + alpha*d */
    double  (*cg_value)  /* user provided routine to return the function */
        (double *, INT), /* value at x */
    void     (*cg_grad)  /* user provided routine, returns in g the */
     (double *, double *, INT), /* gradient at x*/
    cg_com        *Com   /* cg com structure */
)
{
    double a, dphia, b, dphib, c, alpha, phi, dphi,
           a0, da0, b0, db0, width, fquad, zero, rho,
           cg_dot (double *, double *, INT) ;

    INT n, iter ;

    int nsecant, nshrink, ngrow, status, i,
        cg_Wolfe (double, double, double, cg_com *),
        cg_update (double *, double *, double *, double *, double *, double *,
                   double *, double *, double *, double *, double *,
                   double (double *, INT), void (double *, double *, INT),
                   cg_com *) ;

    cg_parameter *Parm ;

    Parm = Com->Parm ;
    alpha = Com->alpha ;
    phi = Com->f ;
    dphi = Com->df ;
    n = Com->n ;
    rho = Parm->rho ;
    zero = 0. ;
    cg_step (xtemp, x, d, alpha, n) ;
    cg_grad (gtemp, xtemp, n) ;
    Com->ng++ ;
    dphi = cg_dot (gtemp, d, n) ;

    /* check if gradient is nan; if so, reduce stepsize */
    if ( dphi != dphi )
    {
        for (i = 0; i < Parm->nexpand; i++)
        {
            alpha /= rho ;
            cg_step (xtemp, x, d, alpha, n) ;
            cg_grad (gtemp, xtemp, n) ;
            Com->ng++ ;
            dphi = cg_dot (gtemp, d, n) ;
            if ( dphi == dphi ) break ;
        }
        if ( i == Parm->nexpand )
        {
            status = -2 ;
            goto Exit ;
        }
        rho = Parm->nan_rho ;
    }
 
/*Find initial interval [a,b] such that dphia < 0, dphib >= 0,
         and phia <= phi0 + feps*fabs (phi0) */
 
    a = zero ;
    dphia = dphi0  ;
    ngrow = 0 ;
    nshrink = 0 ;
    while ( dphi < zero )
    {
        phi = cg_value (xtemp, n) ;
        Com->nf++ ;

/* if quadstep in effect and quadratic conditions hold, check wolfe condition*/

        if ( Com->QuadOK )
        {
            if ( ngrow == 0 ) fquad = MIN (phi, Com->f0) ;
            if ( phi <= fquad )
            {
                if ( Parm->PrintSteps )
                {
                    printf ("alpha: %14.6e phi: %14.6e fquad: %14.6e\n",
                            alpha, phi, fquad) ;
                }
                if ( cg_Wolfe (alpha, phi, dphi, Com) )
                {
                    status = 0 ;
                    goto Exit ;
                }
            }
        }
        if ( phi > Com->fpert )
        {
            /* contraction phase, only break at termination or Secant step */
            b = alpha ;
            while ( TRUE )
            {
                alpha = .5*(a+b) ;
                nshrink++ ;
                if ( nshrink > Parm->nexpand )
                {
                    status = 6 ;
                    goto Exit ;
                }
                cg_step (xtemp, x, d, alpha, n) ;
                cg_grad (gtemp, xtemp, n) ;
                Com->ng++ ;
                dphi = cg_dot (gtemp, d, n) ;
                if ( dphi >= zero ) goto Secant ;
                phi = cg_value (xtemp, n) ;
                Com->nf++ ;
                if ( Parm->PrintSteps )
                {
                    printf ("contract, a: %14.6e b: %14.6e alpha: %14.6e phi: "
                            "%14.6e dphi: %14.6e\n", a, b, alpha, phi, dphi) ;
                }
                if ( Com->QuadOK && (phi <= fquad) )
                {
                    if ( cg_Wolfe (alpha, phi, dphi, Com) )
                    {
                        status = 0 ;
                        goto Exit ;
                    }
                }
                if ( phi <= Com->fpert )
                {
                    a = alpha ;
                    dphia = dphi ;
                }
                else
                {
                    b = alpha ;
                }
            }
        }

/* expansion phase */

        a = alpha ;
        dphia = dphi ;
        ngrow++ ;
        if ( ngrow > Parm->nexpand )
        {
            status = 3 ;
            goto Exit ;
        }
        alpha = rho*alpha ;
        cg_step (xtemp, x, d, alpha, n) ;
        cg_grad (gtemp, xtemp, n) ;
        Com->ng++ ;
        dphi = cg_dot (gtemp, d, n) ;
        if ( Parm->PrintSteps )
        {
            printf ("expand,   a: %14.6e alpha: %14.6e phi: "
                     "%14.6e dphi: %14.6e\n", a, alpha, phi, dphi) ;
        }
    }

Secant:
    b = alpha ;
    dphib = dphi ;
    if ( Com->QuadOK )
    {
        phi = cg_value (xtemp, n) ;
        Com->nf++ ;
        if ( ngrow + nshrink == 0 ) fquad = MIN (phi, Com->f0) ;
        if ( phi <= fquad )
        {
            if ( cg_Wolfe (alpha, phi, dphi, Com) )
            {
                status = 0 ;
                goto Exit ;
            }
        }
    }
    nsecant = Parm->nsecant ;
    for (iter = 1; iter <= nsecant; iter++)
    {
        if ( Parm->PrintSteps )
        {
            printf ("secant, a: %14.6e b: %14.6e da: %14.6e db: %14.6e\n",
                     a, b, dphia, dphib) ;
        }
        width = Parm->gamma*(b - a) ;
        if ( -dphia <= dphib ) alpha = a - (a-b)*(dphia/(dphia-dphib)) ;
        else                   alpha = b - (a-b)*(dphib/(dphia-dphib)) ;
        c = alpha ;
        a0 = a ;
        b0 = b ;
        da0 = dphia ;
        db0 = dphib ;
        status = cg_update (&a, &dphia, &b, &dphib, &alpha, &phi,
                    &dphi, x, xtemp, d, gtemp, cg_value, cg_grad, Com) ;
        if ( status >= 0 ) goto Exit ;
        else if ( status == -2 )
        {
            if ( c == a )
            {
                if ( dphi > da0 ) alpha = c - (c-a0)*(dphi/(dphi-da0)) ;
                else              alpha = a ;
            }
            else
            {
                if ( dphi < db0 ) alpha = c - (c-b0)*(dphi/(dphi-db0)) ;
                else              alpha = b ;
            }
            if ( (alpha > a) && (alpha < b) )
            {
                if ( Parm->PrintSteps ) printf ("2nd secant\n") ;
                status = cg_update (&a, &dphia, &b, &dphib, &alpha, &phi,
                           &dphi, x, xtemp, d, gtemp, cg_value, cg_grad, Com) ;
                if ( status >= 0 ) goto Exit ;
            }
        }

/* bisection iteration */

        if ( b-a >= width )
        {
            alpha = .5*(b+a) ;
            if ( Parm->PrintSteps ) printf ("bisection\n") ;
            status = cg_update (&a, &dphia, &b, &dphib, &alpha, &phi,
                        &dphi, x, xtemp, d, gtemp, cg_value, cg_grad, Com) ;
            if ( status >= 0 ) goto Exit ;
        }
        else if ( b <= a )
        {
            status = 7 ;
            goto Exit ;
        }
    }
    status = 4 ;

Exit:
    Com->alpha = alpha ;
    Com->f = phi ;
    Com->df = dphi ;
    return (status) ;
}

/*
   ordinary Wolfe line search routine
   This routine is identical to cg_line except that the function
   psi [a] = phi [a] - phi [0] - a*delta*dphi [0] is minimized instead of
   the function phi
*/

int cg_lineW
(
    double       dphi0 , /* function derivative at starting point (alpha = 0) */
    double          *x , /* current iterate */
    double      *xtemp , /* x + alpha*d */
    double          *d , /* current search direction */
    double      *gtemp , /* gradient at x + alpha*d */
    double  (*cg_value)  /* user provided routine to return the function */
        (double *, INT), /* value at x */
    void     (*cg_grad)  /* user provided routine, returns in g the */
    (double *, double*, INT), /* gradient at x*/
    cg_com        *Com   /* cg com structure */
)
{
    double a, dpsia, b, dpsib, c, alpha, phi, dphi,
           a0, da0, b0, db0, width, fquad, zero, rho, psi, dpsi,
           cg_dot (double *, double *, INT) ;

    INT n, iter ;

    int nsecant, nshrink, ngrow, status, i,
        cg_Wolfe (double, double, double, cg_com *),
        cg_updateW (double *, double *, double *, double *, double *, double *,
                    double *, double *, double *, double *, double *, double *,
                    double (double *, INT), void (double *, double *, INT),
                    cg_com *) ;
    cg_parameter *Parm ;

    Parm = Com->Parm ;
    alpha = Com->alpha ;
    phi = Com->f ;
    dphi = Com->df ;
    n = Com->n ;
    rho = Parm->rho ;
    zero = 0. ;
    cg_step (xtemp, x, d, alpha, n) ;
    cg_grad (gtemp, xtemp, n) ;
    Com->ng++ ;
    dphi = cg_dot (gtemp, d, n) ;

    /* check if gradient is nan; if so, reduce stepsize */
    if ( dphi != dphi )
    {
        for (i = 0; i < Parm->nexpand; i++)
        {
            alpha /= rho ;
            cg_step (xtemp, x, d, alpha, n) ;
            cg_grad (gtemp, xtemp, n) ;
            Com->ng++ ;
            dphi = cg_dot (gtemp, d, n) ;
            if ( dphi == dphi ) break ;
        }
        if ( i == Parm->nexpand )
        {
            status = -2 ;
            goto Exit ;
        }
        rho = Parm->nan_rho ;
    }
    dpsi = dphi - Com->wolfe_hi ;
 
/*Find initial interval [a,b] such that dphia < 0, dphib >= 0,
         and phia <= phi0 + feps*fabs (phi0) */
 
    a = zero ;
    dpsia = dphi0 - Com->wolfe_hi ;
    ngrow = 0 ;
    nshrink = 0 ;
    while ( dpsi < zero )
    {
        phi = cg_value (xtemp, n) ;
        psi = phi - alpha*Com->wolfe_hi ;
        Com->nf++ ;

/* if quadstep in effect and quadratic conditions hold, check Wolfe condition*/

        if ( Com->QuadOK )
        {
            if ( ngrow == 0 ) fquad = MIN (phi, Com->f0) ;
            if ( phi <= fquad )
            {
                if ( Parm->PrintSteps )
                {
                    printf ("alpha: %14.6e phi: %14.6e fquad: %14.6e\n",
                            alpha, phi, fquad) ;
                }
                if ( cg_Wolfe (alpha, phi, dphi, Com) )
                {
                    status = 0 ;
                    goto Exit ;
                }
            }
        }
        if ( psi <= Com->fpert )
        {
            a = alpha ;
            dpsia = dphi ;
        }
        else
        {
            /* contraction phase, only break at termination or Secant step */
            b = alpha ;
            while ( TRUE )
            {
                alpha = .5*(a+b) ;
                nshrink++ ;
                if ( nshrink > Parm->nexpand )
                {
                    status = 6 ;
                    goto Exit ;
                }
                cg_step (xtemp, x, d, alpha, n) ;
                cg_grad (gtemp, xtemp, n) ;
                Com->ng++ ;
                dphi = cg_dot (gtemp, d, n) ;
                dpsi = dphi - Com->wolfe_hi ;
                if ( dpsi >= zero ) goto Secant ;
                phi = cg_value (xtemp, n) ;
                psi = phi - alpha*Com->wolfe_hi ;
                Com->nf++ ;
                if ( Parm->PrintSteps )
                {
                    printf ("contract, a: %14.6e b: %14.6e alpha: %14.6e phi: "
                            "%14.6e dphi: %14.6e\n", a, b, alpha, phi, dphi) ;
                }
                if ( Com->QuadOK && (phi <= fquad) )
                {
                    if ( cg_Wolfe (alpha, phi, dphi, Com) )
                    {
                        status = 0 ;
                        goto Exit ;
                    }
                }
                if ( psi <= Com->fpert )
                {
                    a = alpha ;
                    dpsia = dpsi ;
                }
                else
                {
                    b = alpha ;
                }
            }
        }

/* expansion phase */

        ngrow++ ;
        if ( ngrow > Parm->nexpand )
        {
            status = 3 ;
            goto Exit ;
        }
        alpha *= rho ;
        cg_step (xtemp, x, d, alpha, n) ;
        cg_grad (gtemp, xtemp, n) ;
        Com->ng++ ;
        dphi = cg_dot (gtemp, d, n) ;
        dpsi = dphi - Com->wolfe_hi ;
        if ( Parm->PrintSteps )
        {
            printf ("expand,   a: %14.6e alpha: %14.6e phi: "
                     "%14.6e dphi: %14.6e\n", a, alpha, phi, dphi) ;
        }
    }

Secant:
    b = alpha ;
    dpsib = dpsi ;
    if ( Com->QuadOK )
    {
        phi = cg_value (xtemp, n) ;
        Com->nf++ ;
        if ( ngrow + nshrink == 0 ) fquad = MIN (phi, Com->f0) ;
        if ( phi <= fquad )
        {
            if ( cg_Wolfe (alpha, phi, dphi, Com) )
            {
                status = 0 ;
                goto Exit ;
            }
        }
    }
    nsecant = Parm->nsecant ;
    for (iter = 1; iter <= nsecant; iter++)
    {
        if ( Parm->PrintSteps )
        {
            printf ("secant, a: %14.6e b: %14.6e da: %14.6e db: %14.6e\n",
                     a, b, dpsia, dpsib) ;
        }
        width = Parm->gamma*(b - a) ;
        if ( -dpsia <= dpsib ) alpha = a - (a-b)*(dpsia/(dpsia-dpsib)) ;
        else                   alpha = b - (a-b)*(dpsib/(dpsia-dpsib)) ;
        c = alpha ;
        a0 = a ;
        b0 = b ;
        da0 = dpsia ;
        db0 = dpsib ;
        status = cg_updateW (&a, &dpsia, &b, &dpsib, &alpha, &phi, &dphi,
                   &dpsi, x, xtemp, d, gtemp, cg_value, cg_grad, Com) ;
        if ( status >= 0 ) goto Exit ;
        else if ( status == -2 )
        {
            if ( c == a )
            {
                if ( dpsi > da0 ) alpha = c - (c-a0)*(dpsi/(dpsi-da0)) ;
                else              alpha = a ;
            }
            else
            {
                if ( dpsi < db0 ) alpha = c - (c-b0)*(dpsi/(dpsi-db0)) ;
                else              alpha = b ;
            }
            if ( (alpha > a) && (alpha < b) )
            {
                if ( Parm->PrintSteps ) printf ("2nd secant\n") ;
                status = cg_updateW (&a, &dpsia, &b, &dpsib, &alpha, &phi,
                    &dphi, &dpsi, x, xtemp, d, gtemp, cg_value, cg_grad, Com) ;
                if ( status >= 0 ) goto Exit ;
            }
        }

/* bisection iteration */

        if ( b-a >= width )
        {
            alpha = .5*(b+a) ;
            if ( Parm->PrintSteps ) printf ("bisection\n") ;
            status = cg_updateW (&a, &dpsia, &b, &dpsib, &alpha, &phi, &dphi,
                       &dpsi, x, xtemp, d, gtemp, cg_value, cg_grad, Com) ;
            if ( status >= 0 ) goto Exit ;
        }
        else if ( b <= a )
        {
            status = 7 ;
            goto Exit ;
        }
    }
    status = 4 ;

Exit:
    Com->alpha = alpha ;
    Com->f = phi ;
    Com->df = dphi ;
    return (status) ;
}

/* update returns: 8 if too many iterations
                   0 if Wolfe condition is satisfied
                  -1 if interval is updated and a search is done
                  -2 if the interval updated successfully */

int cg_update
(
    double          *a , /* left side of bracketing interval */
    double      *dphia , /* derivative at a */
    double          *b , /* right side of bracketing interval */
    double      *dphib , /* derivative at b */
    double      *alpha , /* trial step (between a and b) */
    double        *phi , /* function value at alpha (returned) */
    double       *dphi , /* function derivative at alpha (returned) */
    double          *x , /* current iterate */
    double      *xtemp , /* x + alpha*d */
    double          *d , /* current search direction */
    double      *gtemp , /* gradient at x + alpha*d */
    double    (*cg_value)  /* user provided routine to return the function */
          (double *, INT), /* value at x */
    void       (*cg_grad)  /* user provided routine, returns in g the */
     (double *, double*, INT), /* gradient at x*/
    cg_com        *Com   /* cg com structure */
)
{
    double zero, cg_dot (double *, double *, INT) ;
    INT n ;
    int nshrink, status, cg_Wolfe (double, double, double, cg_com *) ;
    cg_parameter *Parm ;

    Parm = Com->Parm ;
    zero = 0. ;
    n = Com->n ;
    cg_step (xtemp, x, d, *alpha, n) ;
    *phi = cg_value (xtemp, n) ;
    Com->nf++ ;
    cg_grad (gtemp, xtemp, n) ;
    Com->ng++ ;
    *dphi = cg_dot (gtemp, d, n) ;
    if ( Parm->PrintSteps )
    {
        printf ("update alpha: %14.6e phi: %14.6e dphi: %14.6e\n",
                 *alpha, *phi, *dphi) ;
    }
    if ( cg_Wolfe (*alpha, *phi, *dphi, Com) )
    {
        status = 0 ;
        goto Exit2 ;
    }
    status = -2 ;
    if ( *dphi >= zero )
    {
        *b = *alpha ;
        *dphib = *dphi ;
        goto Exit2 ;
    }
    else
    {
        if ( *phi <= Com->fpert )
        {
            *a = *alpha ;
            *dphia = *dphi ;
            goto Exit2 ;
        }
    }
    nshrink = 0 ;
    *b = *alpha ;
    while ( TRUE )
    {
        *alpha = .5*(*a + *b) ;
        nshrink++ ;
        if ( nshrink > Parm->nexpand )
        {
            status = 8 ;
            goto Exit2 ;
        }
        cg_step (xtemp, x, d, *alpha, n) ;
        cg_grad (gtemp, xtemp, n) ;
        Com->ng++ ;
        *dphi = cg_dot (gtemp, d, n) ;
        *phi = cg_value (xtemp, n) ;
        Com->nf++ ;
        if ( Parm->PrintSteps )
        {
            printf ("contract, a: %14.6e alpha: %14.6e "
                    "phi: %14.6e dphi: %14.6e\n", *a, *alpha, *phi, *dphi) ;
        }
        if ( cg_Wolfe (*alpha, *phi, *dphi, Com) )
        {
            status = 0 ;
            goto Exit2 ;
        }
        if ( *dphi >= zero )
        {
            *b = *alpha ;
            *dphib = *dphi ;
            goto Exit1 ;
        }
        if ( *phi <= Com->fpert )
        {
            if ( Parm->PrintSteps )
            {
                printf ("update a: %14.6e dphia: %14.6e\n", *alpha, *dphi) ;
            }
            *a = *alpha ;
            *dphia = *dphi ;
        }
        else *b = *alpha ;
    }
Exit1:
    status = -1 ;
Exit2:
    if ( Parm->PrintSteps )
    {
        printf ("UP a: %14.6e b: %14.6e da: %14.6e db: %14.6e status: %i\n",
                 *a, *b, *dphia, *dphib, status) ;
    }
    return (status) ;
}

/* This routine is identical to cg_update except that the function
   psi [a] = phi [a] - phi [0] - a*delta*dphi [0] is minimized instead of
   the function phi. The return int has the following meaning:
                   8 if too many iterations
                   0 if Wolfe condition is satisfied
                  -1 if interval is updated and a search is done
                  -2 if the interval updated successfully */

int cg_updateW
(
    double          *a , /* left side of bracketing interval */
    double      *dpsia , /* derivative at a */
    double          *b , /* right side of bracketing interval */
    double      *dpsib , /* derivative at b */
    double      *alpha , /* trial step (between a and b) */
    double        *phi , /* function value at alpha (returned) */
    double       *dphi , /* derivative of phi at alpha (returned) */
    double       *dpsi , /* derivative of psi at alpha (returned) */
    double          *x , /* current iterate */
    double      *xtemp , /* x + alpha*d */
    double          *d , /* current search direction */
    double      *gtemp , /* gradient at x + alpha*d */
    double (*cg_value)   /* user provided routine to return the function */
       (double *, INT),  /* value at x */
    void    (*cg_grad)   /* user provided routine, returns in g the */
     (double *, double*, INT), /* gradient at x*/
    cg_com        *Com   /* cg com structure */
)
{
    double zero, psi, cg_dot (double *, double *, INT) ;
    INT n ;
    int nshrink, status, cg_Wolfe (double, double, double, cg_com *) ;
    cg_parameter *Parm ;

    Parm = Com->Parm ;
    zero = 0. ;
    n = Com->n ;
    cg_step (xtemp, x, d, *alpha, n) ;
    *phi = cg_value (xtemp, n) ;
    psi = *phi - *alpha*Com->wolfe_hi ;
    Com->nf++ ;
    cg_grad (gtemp, xtemp, n) ;
    Com->ng++ ;
    *dphi = cg_dot (gtemp, d, n) ;
    *dpsi = *dphi - Com->wolfe_hi ;
    if ( Parm->PrintSteps )
    {
        printf ("update alpha: %14.6e psi: %14.6e dpsi: %14.6e\n",
                 *alpha, psi, *dpsi) ;
    }
    if ( cg_Wolfe (*alpha, *phi, *dphi, Com) )
    {
        status = 0 ;
        goto Exit2 ;
    }
    status = -2 ;
    if ( *dpsi >= zero )
    {
        *b = *alpha ;
        *dpsib = *dpsi ;
        goto Exit2 ;
    }
    else
    {
        if ( psi <= Com->fpert )
        {
            *a = *alpha ;
            *dpsia = *dpsi ;
            goto Exit2 ;
        }
    }
    nshrink = 0 ;
    *b = *alpha ;
    while ( TRUE )
    {
        *alpha = .5*(*a + *b) ;
        nshrink++ ;
        if ( nshrink > Parm->nexpand )
        {
            status = 8 ;
            goto Exit2 ;
        }
        cg_step (xtemp, x, d, *alpha, n) ;
        cg_grad (gtemp, xtemp, n) ;
        Com->ng++ ;
        *dphi = cg_dot (gtemp, d, n) ;
        *dpsi = *dphi - Com->wolfe_hi ;
        *phi = cg_value (xtemp, n) ;
        psi = *phi - *alpha*Com->wolfe_hi ;
        Com->nf++ ;
        if ( Parm->PrintSteps )
        {
            printf ("contract, a: %14.6e alpha: %14.6e "
                    "phi: %14.6e dphi: %14.6e\n", *a, *alpha, *phi, *dphi) ;
        }
        if ( cg_Wolfe (*alpha, *phi, *dphi, Com) )
        {
            status = 0 ;
            goto Exit2 ;
        }
        if ( *dpsi >= zero )
        {
            *b = *alpha ;
            *dpsib = *dpsi ;
            goto Exit1 ;
        }
        if ( psi <= Com->fpert )
        {
            if ( Parm->PrintSteps )
            {
                printf ("update a: %14.6e dpsia: %14.6e\n", *alpha, *dpsi) ;
            }
            *a = *alpha ;
            *dpsia = *dpsi ;
        }
        else *b = *alpha ;
    }
Exit1:
    status = -1 ;
Exit2:
    if ( Parm->PrintSteps )
    {
        printf ("UP a: %14.6e b: %14.6e da: %14.6e db: %14.6e status: %i\n",
                 *a, *b, *dpsia, *dpsib, status) ;
    }
    return (status) ;
}

/*
Version 1.2 Change:
  1. The variable dpsi needs to be included in the argument list for
     subroutine cg_updateW (update of a Wolfe line search)

Version 2.0 Changes:
    The user interface was redesigned. The parameters no longer need to
    be read from a file. For compatibility with earlier versions of the
    code, we include the routine cg_readParms to read parameters.
    In the simplest case, the user can use NULL for the
    parameter argument of cg_descent, and the code sets the default
    parameter values. If the user wishes to modify the parameters, call
    cg_default in the main program to initialize a cg_parameter
    structure. Individual elements of the structure could be modified.
    The header file cg_user.h contains the structures and prototypes
    that the user may need to reference or modify, while cg_descent.h
    contains header elements that only cg_descent will access.  Note
    that the arguments of cg_descent have changed.
*/
