#include <limits.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#define INT long int
#define INT_INF LONG_MAX
#define INF DBL_MAX
#define TRUE 1
#define FALSE 0
#define CGMAXLINE 200
#ifndef NULL
#define NULL 0
#endif

typedef struct cg_parameter_struct /* user controlled parameters */
{
    double           delta ; /* Wolfe line search parameter */
    double           sigma ; /* Wolfe line search parameter */
    double           gamma ; /* decay factor for bracket interval width */
    double             rho ; /* growth factor when searching for initial
                                bracketing interval */
    double         nan_rho ; /* after encountering nan, growth factor when
                                searching for initial bracketing interval */
    double             eta ; /* lower bound for the conjugate gradient update
                                parameter beta_k is eta*||d||_2 */
    double             eps ; /* perturbation parameter for computing fpert */
    double          Qdecay ; /* factor used to compute average cost */
    double      QuadCutOff ; /* QuadStep used when relative change in f >
                                QuadCutOff */
    double         StopFac ; /* factor multiplying starting |grad|_infty in
                                the stopping condition StopRule (see below) */
    double       AWolfeFac ; /* if AWolfe is F, then set AWolfe = T when
                                |f-f0| < Awolfe_fac*Ck) */
    double     restart_fac ; /* the conjugate gradient algorithm is restarted
                                every restart_fac*n  iterations */
    double            feps ; /* stop when cost change <= feps*|f| */
    double            psi0 ; /* factor used in starting guess for iteration 1 */
    double            psi1 ; /* in performing a QuadStep, we evaluate the
                                function at psi1*previous step */
    double            psi2 ; /* when starting a new cg iteration, our initial
                                guess for the line search stepsize is
                                psi2*previous step */
    double            step ; /* if nonzero, it is the initial step of initial
                                line search */
    int            nexpand ; /* abort the line search if the bracketing
                                interval is expanded nexpand times */
    int            nsecant ; /* abort the line search if the number of secant
                                iterations is nsecant */
    double       maxit_fac ; /* abort when number of iterations reaches
                                maxit_fac*n */

    /* user controlled logical parameters, T means TRUE, F means FALSE */
    int           PertRule ; /* error estimate for function value,
                                F => eps, T => eps*Ck */
    int           QuadStep ; /* T (use quadratic interpolation in line search)*/
    int         PrintFinal ; /* T (print final function value) F (no printing)*/
    int         PrintSteps ; /* T (print each step) F (no print) */
    int         PrintParms ; /* T (print parameters) F (do not display Parms) */
    int           StopRule ; /* T (|grad|_infty <=
                                   max(tol, initial |grad|_infty*StopFact)
                                F (|grad|_infty <= tol*(1+|f|)) */
    int             AWolfe ; /* T (use approximate Wolfe line search)
                                F (use Wolfe line search) */
    int              debug ; /* T (check for no increase in function value)
                                F (no debugging) */
} cg_parameter ;

typedef struct cg_stats_struct /* statistics returned to user */
{
    double               f ; /*function value at solution */
    double           gnorm ; /* max abs component of gradient */
    INT               iter ; /* number of iterations */
    INT              nfunc ; /* number of function evaluations */
    INT              ngrad ; /* number of gradient evaluations */
} cg_stats ;

/* prototypes */
int cg_descent /*  return:
                           0 (convergence tolerance satisfied)
                           1 (change in func <= feps*|f|)
                           2 (total iterations exceeded maxit)
                           3 (slope always negative in line search)
                           4 (number secant iterations exceed nsecant)
                           5 (search direction not a descent direction)
                           6 (line search fails in initial interval)
                           7 (line search fails during bisection)
                           8 (line search fails during interval update)
                           9 (debugger is on and the function value increases)*/
(
    double            *x , /* input: starting guess, output: the solution */
    INT                n , /* problem dimension */
    cg_stats      *Stats , /* structure with statistics (see cg_descent.h) */
    cg_parameter  *UParm , /* user parameters, NULL = use default parameters */
    double      grad_tol , /* StopRule = 1: |g|_infty <= max (grad_tol,
                                            StopFac*initial |g|_infty) [default]
                              StopRule = 0: |g|_infty <= grad_tol(1+|f|) */
    double    (*cg_value)  /* user provided routine to return the function */
          (double *, INT), /* value cg_value(x) at x */
    void       (*cg_grad)  /* user provided routine cg_grad (g, x), g is*/
     (double *, double *, INT), /* the gradient at x*/
    double         *work   /* work array with at least 4n elements */
) ;

void cg_default /* set default parameter values */
(
    cg_parameter   *Parm
) ;

int  cg_readParms /* return:
                             0 (parameter file was read)
                            -1 (parameter file not found)
                            -2 (missing entry in parameter file)
                            -3 (comment in parameter file too long) */
(
    char        *filename ,
    cg_parameter    *Parm
) ;
