/* The operation of the code is mostly controlled by the parameters
   in the cg_parameter structure.  In the following example,
   the parameter QuadStep is set to FALSE.  When QuadStep is TRUE,
   the trial step in each iteration is computed as the minimizer
   of a quadratic interpolant along the search direction. In
   performing the quadstep, we hope to find a suitable line search
   point right away, completely by-passing the secant iteration.
   However, as the iterates approach a minimizer, the numerical accuracy of
   the minimizer of the quadratic interpolant becomes worse. When the relative
   change in the function values for two consecutive iterations reach
   QuadCutOff, then the code completely turns off the quadstep. The user
   can turn off the quadstep by setting QuadStep FALSE. By leaving
   QuadStep TRUE, but increasing QuadCutOff (default 1.d-12), the code turns
   off the QuadStep sooner. Note that the code is usually faster when the
   quadstep is employed. For the test problem, turning off QuadStep
   (setting its value to FALSE) yields the following output.
   Notice that the number of gradient evaluations increases while the
   number of function evaluations decreases (compared to the previous run).
   This behavior is typical.

    Termination status: 0
    Convergence tolerance for gradient satisfied
    absolute largest component of gradient: 5.950597e-09
    function value: -6.530787e+02
    cg iterations: 35
    function evaluations: 39
    gradient evaluations: 74 */

#include <math.h>
#include "cg_user.h"

double myvalue
(
    double   *x ,
    INT       n
) ;

void mygrad
(
    double    *g ,
    double    *x ,
    INT        n
) ;

int main (void)
{
    double *x, *work ;
    INT i, mydim ;
    int status ;
    cg_stats Stats ;
    cg_parameter Parm ;

    cg_default (&Parm) ;
    Parm.QuadStep = FALSE ;
/*  Parm.PrintParms = TRUE ;*/

/* allocate workspace and set parameters */
    mydim = 100 ;
    x = (double *) malloc (mydim*sizeof (double)) ;
    work = (double *) malloc (4*mydim*sizeof (double)) ;

/* starting guess */
    for (i = 0; i < mydim; i++) x [i] = 1. ;

/* run the code */
    status = cg_descent (x, mydim, &Stats, &Parm, 1.e-8, myvalue, mygrad, work);

/* free workspace */
    free (x) ;
    free (work) ;
}

double myvalue
(
    double   *x ,
    INT       n
)
{
    double f, t ;
    INT i ;
    f = 0. ;
    for (i = 0; i < n; i++)
    {
        t = i+1 ;
        t = sqrt (t) ;
        f += exp (x [i]) - t*x [i] ;
    }
    return (f) ;
}

void mygrad
(
    double    *g ,
    double    *x ,
    INT        n
)
{
    double t ;
    INT i ;
    for (i = 0; i < n; i++)
    {
        t = i + 1 ;
        t = sqrt (t) ;
        g [i] = exp (x [i]) -  t ;
    }
    return ;
}
