/* Although there is a rigorous theory justifying a Wolfe line search,
   the performance of the Approximate Wolfe line search is often superior.
   Nonetheless, the user can completely turn off the Approximate
   Wolfe line search by setting AWolfe to FALSE and AWolfeFac to 0.
   The output is the following:

   Termination status: 4
   Line search fails, too many secant steps
   Possible causes of this error message:
      - your tolerance may be too strict: grad_tol =  1.000000e-08
   absolute largest component of gradient: 1.562251e-07
   function value: -6.530787e+02
   cg iterations: 29
   function evaluations: 150
   gradient evaluations: 137

   Hence, due to numerical errors, it was not possible to achieve the specified
   1.e-8 error tolerance. By decreasing the error tolerance to 1.e-6,
   the problem is solved:

   Termination status: 0
   Convergence tolerance for gradient satisfied
   absolute largest component of gradient: 7.846751e-07
   function value: -6.530787e+02
   cg iterations: 26
   function evaluations: 49
   gradient evaluations: 33

   On the other hand, if we turn on the Approximate Wolfe line search by
   resetting AWolfeFac to its default value 1.e-3, we
   obtain slightly faster convergence:

   Termination status: 0
   Convergence tolerance for gradient satisfied
   absolute largest component of gradient: 9.693105e-07
   function value: -6.530787e+02
   cg iterations: 25
   function evaluations: 48
   gradient evaluations: 31 */

#include <math.h>
#include "cg_user.h"

double myvalue
(
    double   *x ,
    INT       n
) ;

void mygrad
(
    double    *g ,
    double    *x ,
    INT        n
) ;

int main (void)
{
    double *x, *work ;
    INT i, mydim ;
    int status ;
    cg_stats Stats ;
    cg_parameter Parm ;

/* set parameter values */
    cg_default (&Parm) ;
    Parm.AWolfe = FALSE ;
    Parm.AWolfeFac = 0.e-3 ;

/* allocate work space */
    mydim = 100 ;
    x = (double *) malloc (mydim*sizeof (double)) ;
    work = (double *) malloc (4*mydim*sizeof (double)) ;

/* starting guess */
    for (i = 0; i < mydim; i++) x [i] = 1. ;

/* solve the problem */
    status = cg_descent (x, mydim, &Stats, &Parm, 1.e-8, myvalue, mygrad, work);

/* free work space */
    free (x) ;
    free (work) ;
}

double myvalue
(
    double   *x ,
    INT       n
)
{
    double f, t ;
    INT i ;
    f = 0. ;
    for (i = 0; i < n; i++)
    {
        t = i+1 ;
        t = sqrt (t) ;
        f += exp (x [i]) - t*x [i] ;
    }
    return (f) ;
}

void mygrad
(
    double    *g ,
    double    *x ,
    INT        n
)
{
    double t ;
    INT i ;
    for (i = 0; i < n; i++)
    {
        t = i + 1 ;
        t = sqrt (t) ;
        g [i] = exp (x [i]) -  t ;
    }
    return ;
}
